@@ -2,7 +2,7 @@
 //! Run Actix Web on AWS Lambda
 //!
 //!
-use crate::request::ApiGatewayV2;
+use crate::request::LambdaHttpEvent;
 use core::convert::TryFrom;
 use core::future::Future;
 use lambda_runtime::{
@@ -75,7 +75,7 @@ where
     B: actix_web::body::MessageBody,
     B::Error: std::fmt::Display;
 
-impl<S, B> LambdaHandler<ApiGatewayV2<'_>, serde_json::Value> for ActixHandler<S, B>
+impl<S, B> LambdaHandler<LambdaHttpEvent<'_>, serde_json::Value> for ActixHandler<S, B>
 where
     S: actix_service::Service<
             actix_http::Request,
@@ -91,11 +91,13 @@ where
     /// Lambda handler function
     /// Parse Lambda event as Actix-web request,
     /// serialize Actix-web response to Lambda JSON response
-    fn call(&self, event: ApiGatewayV2, _context: LambdaContext) -> Self::Fut {
+    fn call(&self, event: LambdaHttpEvent, _context: LambdaContext) -> Self::Fut {
         use serde_json::json;
 
         // check if web client supports content-encoding: br
-        let client_br = crate::brotli::client_supports_brotli(&event);
+        let client_br = event.client_supports_brotli();
+        // multi-value-headers response format
+        let multi_value = event.multi_value();
 
         // Parse request
         let actix_request = actix_http::Request::try_from(event);
@@ -109,7 +111,7 @@ where
                     // Request parsing succeeded
                     if let Ok(response) = svc_fut.await {
                         // Returns as API Gateway response
-                        api_gateway_response_from_actix_web(response, client_br).await
+                        api_gateway_response_from_actix_web(response, client_br, multi_value).await
                     } else {
                         // Some Actix web error -> 500 Internal Server Error
                         Ok(json!({
@@ -135,64 +137,32 @@ where
     }
 }
 
-impl TryFrom<ApiGatewayV2<'_>> for actix_http::Request {
+impl TryFrom<LambdaHttpEvent<'_>> for actix_http::Request {
     type Error = LambdaError;
 
     /// Actix-web Request from API Gateway event
-    fn try_from(event: ApiGatewayV2) -> Result<Self, Self::Error> {
-        use actix_web::cookie::Cookie;
+    fn try_from(event: LambdaHttpEvent) -> Result<Self, Self::Error> {
         use actix_web::http::Method;
-        use std::borrow::Cow;
-        use std::net::IpAddr;
-        use std::str::FromStr;
-
-        // path ? query_string
-        let path_and_query: Cow<str> = if event.raw_query_string.is_empty() {
-            event.encoded_path()
-        } else {
-            format!("{}?{}", event.encoded_path(), event.raw_query_string).into()
-        };
-
-        // Method, Source IP
-        let method = Method::try_from(&event.request_context.http.method as &str)?;
-        let source_ip = IpAddr::from_str(&event.request_context.http.source_ip as &str)?;
 
         // Construct actix_web request
-        let req = actix_web::test::TestRequest::with_uri(&path_and_query)
-            .method(method)
-            .peer_addr(std::net::SocketAddr::from((source_ip, 0u16)));
+        let method = Method::try_from(event.method())?;
+        let req = actix_web::test::TestRequest::with_uri(&event.path_query()).method(method);
 
-        // Cookies
-        let req = if let Some(cookies) = event.cookies {
-            cookies.iter().fold(req, |req, cookie| {
-                if let Ok(cookie_decoded) = Cookie::parse_encoded(cookie as &str) {
-                    req.cookie(cookie_decoded)
-                } else {
-                    req
-                }
-            })
+        // Source IP
+        let req = if let Some(source_ip) = event.source_ip() {
+            req.peer_addr(std::net::SocketAddr::from((source_ip, 0u16)))
         } else {
             req
         };
 
         // Headers
         let req = event
-            .headers
-            .iter()
-            .fold(req, |req, (k, v)| req.insert_header((k as &str, v as &str)));
+            .headers()
+            .into_iter()
+            .fold(req, |req, (k, v)| req.insert_header((k, &v as &str)));
 
         // Body
-        let req = if let Some(eventbody) = event.body {
-            if event.is_base64_encoded {
-                // base64 decode
-                let binarybody = base64::decode(&eventbody as &str)?;
-                req.set_payload(binarybody)
-            } else {
-                req.set_payload((&eventbody as &str).to_string())
-            }
-        } else {
-            req
-        };
+        let req = req.set_payload(event.body()?);
 
         Ok(req.to_request())
     }
@@ -218,18 +188,37 @@ impl<B> crate::brotli::ResponseCompression for actix_web::dev::ServiceResponse<B
 async fn api_gateway_response_from_actix_web<B: actix_web::body::MessageBody>(
     response: actix_web::dev::ServiceResponse<B>,
     client_support_br: bool,
+    multi_value: bool,
 ) -> Result<serde_json::Value, B::Error> {
     use crate::brotli::ResponseCompression;
+    use actix_web::http::header::SET_COOKIE;
     use serde_json::json;
 
     // HTTP status
     let status_code = response.status().as_u16();
 
     // Convert header to JSON map
+    let mut cookies = Vec::<String>::new();
     let mut headers = serde_json::Map::new();
     for (k, v) in response.headers() {
         if let Ok(value_str) = v.to_str() {
-            headers.insert(k.as_str().to_string(), json!(value_str));
+            if multi_value {
+                // REST API format, returns multiValueHeaders
+                if let Some(values) = headers.get_mut(k.as_str()) {
+                    if let Some(value_ary) = values.as_array_mut() {
+                        value_ary.push(json!(value_str));
+                    }
+                } else {
+                    headers.insert(k.as_str().to_string(), json!([value_str]));
+                }
+            } else {
+                // HTTP API v2 format, returns headers
+                if k == SET_COOKIE {
+                    cookies.push(value_str.to_string());
+                } else {
+                    headers.insert(k.as_str().to_string(), json!(value_str));
+                }
+            }
         }
     }
 
@@ -237,79 +226,107 @@ async fn api_gateway_response_from_actix_web<B: actix_web::body::MessageBody>(
     let compress = client_support_br && response.can_brotli_compress();
     let body_bytes = actix_web::body::to_bytes(response.into_body()).await?;
     let body_base64 = if compress {
-        headers.insert("content-encoding".to_string(), json!("br"));
+        if multi_value {
+            headers.insert("content-encoding".to_string(), json!(["br"]));
+        } else {
+            headers.insert("content-encoding".to_string(), json!("br"));
+        }
         crate::brotli::compress_response_body(&body_bytes)
     } else {
         base64::encode(body_bytes)
     };
 
-    Ok(json!({
-        "isBase64Encoded": true,
-        "statusCode": status_code,
-        "headers": headers,
-        "body": body_base64
-    }))
+    if multi_value {
+        Ok(json!({
+            "isBase64Encoded": true,
+            "statusCode": status_code,
+            "multiValueHeaders": headers,
+            "body": body_base64
+        }))
+    } else {
+        Ok(json!({
+            "isBase64Encoded": true,
+            "statusCode": status_code,
+            "cookies": cookies,
+            "headers": headers,
+            "body": body_base64
+        }))
+    }
 }
 
 #[cfg(test)]
 mod tests {
     use super::*;
-    use crate::{request::ApiGatewayV2, test_consts::*};
+    use crate::{request::LambdaHttpEvent, test_consts::*};
+
+    // Request JSON to actix_http::Request
+    fn prepare_request(event_str: &str) -> actix_http::Request {
+        let reqjson: LambdaHttpEvent = serde_json::from_str(event_str).unwrap();
+        actix_http::Request::try_from(reqjson).unwrap()
+    }
 
     #[test]
     fn test_path_decode() {
-        let reqjson: ApiGatewayV2 = serde_json::from_str(API_GATEWAY_V2_GET_ROOT_NOQUERY).unwrap();
-        let req = actix_http::Request::try_from(reqjson).unwrap();
+        let req = prepare_request(API_GATEWAY_V2_GET_ROOT_NOQUERY);
         assert_eq!(req.uri().path(), "/");
+        let req = prepare_request(API_GATEWAY_REST_GET_ROOT_NOQUERY);
+        assert_eq!(req.uri().path(), "/stage/");
 
-        let reqjson: ApiGatewayV2 =
-            serde_json::from_str(API_GATEWAY_V2_GET_SOMEWHERE_NOQUERY).unwrap();
-        let req = actix_http::Request::try_from(reqjson).unwrap();
+        let req = prepare_request(API_GATEWAY_V2_GET_SOMEWHERE_NOQUERY);
         assert_eq!(req.uri().path(), "/somewhere");
+        let req = prepare_request(API_GATEWAY_REST_GET_SOMEWHERE_NOQUERY);
+        assert_eq!(req.uri().path(), "/stage/somewhere");
 
-        let reqjson: ApiGatewayV2 =
-            serde_json::from_str(API_GATEWAY_V2_GET_SPACEPATH_NOQUERY).unwrap();
-        let req = actix_http::Request::try_from(reqjson).unwrap();
+        let req = prepare_request(API_GATEWAY_V2_GET_SPACEPATH_NOQUERY);
         assert_eq!(req.uri().path(), "/path%20with/space");
+        let req = prepare_request(API_GATEWAY_REST_GET_SPACEPATH_NOQUERY);
+        assert_eq!(req.uri().path(), "/stage/path%20with/space");
 
-        let reqjson: ApiGatewayV2 =
-            serde_json::from_str(API_GATEWAY_V2_GET_PERCENTPATH_NOQUERY).unwrap();
-        let req = actix_http::Request::try_from(reqjson).unwrap();
+        let req = prepare_request(API_GATEWAY_V2_GET_PERCENTPATH_NOQUERY);
         assert_eq!(req.uri().path(), "/path%25with/percent");
+        let req = prepare_request(API_GATEWAY_REST_GET_PERCENTPATH_NOQUERY);
+        assert_eq!(req.uri().path(), "/stage/path%25with/percent");
 
-        let reqjson: ApiGatewayV2 =
-            serde_json::from_str(API_GATEWAY_V2_GET_UTF8PATH_NOQUERY).unwrap();
-        let req = actix_http::Request::try_from(reqjson).unwrap();
+        let req = prepare_request(API_GATEWAY_V2_GET_UTF8PATH_NOQUERY);
         assert_eq!(
             req.uri().path(),
             "/%E6%97%A5%E6%9C%AC%E8%AA%9E/%E3%83%95%E3%82%A1%E3%82%A4%E3%83%AB%E5%90%8D"
         );
+        let req = prepare_request(API_GATEWAY_REST_GET_UTF8PATH_NOQUERY);
+        assert_eq!(
+            req.uri().path(),
+            "/stage/%E6%97%A5%E6%9C%AC%E8%AA%9E/%E3%83%95%E3%82%A1%E3%82%A4%E3%83%AB%E5%90%8D"
+        );
     }
 
     #[test]
     fn test_query_decode() {
-        let reqjson: ApiGatewayV2 = serde_json::from_str(API_GATEWAY_V2_GET_ROOT_ONEQUERY).unwrap();
-        let req = actix_http::Request::try_from(reqjson).unwrap();
+        let req = prepare_request(API_GATEWAY_V2_GET_ROOT_ONEQUERY);
+        assert_eq!(req.uri().query(), Some("key=value"));
+        let req = prepare_request(API_GATEWAY_REST_GET_ROOT_ONEQUERY);
         assert_eq!(req.uri().query(), Some("key=value"));
 
-        let reqjson: ApiGatewayV2 =
-            serde_json::from_str(API_GATEWAY_V2_GET_SOMEWHERE_ONEQUERY).unwrap();
-        let req = actix_http::Request::try_from(reqjson).unwrap();
+        let req = prepare_request(API_GATEWAY_V2_GET_SOMEWHERE_ONEQUERY);
+        assert_eq!(req.uri().query(), Some("key=value"));
+        let req = prepare_request(API_GATEWAY_REST_GET_SOMEWHERE_ONEQUERY);
         assert_eq!(req.uri().query(), Some("key=value"));
 
-        let reqjson: ApiGatewayV2 =
-            serde_json::from_str(API_GATEWAY_V2_GET_SOMEWHERE_TWOQUERY).unwrap();
-        let req = actix_http::Request::try_from(reqjson).unwrap();
+        let req = prepare_request(API_GATEWAY_V2_GET_SOMEWHERE_TWOQUERY);
         assert_eq!(req.uri().query(), Some("key1=value1&key2=value2"));
+        let req = prepare_request(API_GATEWAY_REST_GET_SOMEWHERE_TWOQUERY);
+        assert!(
+            req.uri().query() == Some("key1=value1&key2=value2")
+                || req.uri().query() == Some("key2=value2&key1=value1")
+        );
 
-        let reqjson: ApiGatewayV2 =
-            serde_json::from_str(API_GATEWAY_V2_GET_SOMEWHERE_SPACEQUERY).unwrap();
-        let req = actix_http::Request::try_from(reqjson).unwrap();
+        let req = prepare_request(API_GATEWAY_V2_GET_SOMEWHERE_SPACEQUERY);
         assert_eq!(req.uri().query(), Some("key=value1+value2"));
+        let req = prepare_request(API_GATEWAY_REST_GET_SOMEWHERE_SPACEQUERY);
+        assert_eq!(req.uri().query(), Some("key=value1%20value2"));
 
-        let reqjson: ApiGatewayV2 =
-            serde_json::from_str(API_GATEWAY_V2_GET_SOMEWHERE_UTF8QUERY).unwrap();
-        let req = actix_http::Request::try_from(reqjson).unwrap();
+        let req = prepare_request(API_GATEWAY_V2_GET_SOMEWHERE_UTF8QUERY);
+        assert_eq!(req.uri().query(), Some("key=%E6%97%A5%E6%9C%AC%E8%AA%9E"));
+        let req = prepare_request(API_GATEWAY_REST_GET_SOMEWHERE_UTF8QUERY);
         assert_eq!(req.uri().query(), Some("key=%E6%97%A5%E6%9C%AC%E8%AA%9E"));
     }
 
@@ -318,15 +335,23 @@ mod tests {
         use std::net::IpAddr;
         use std::str::FromStr;
 
-        let reqjson: ApiGatewayV2 = serde_json::from_str(API_GATEWAY_V2_GET_ROOT_ONEQUERY).unwrap();
-        let req = actix_http::Request::try_from(reqjson).unwrap();
+        let req = prepare_request(API_GATEWAY_V2_GET_ROOT_ONEQUERY);
+        assert_eq!(
+            req.peer_addr().unwrap().ip(),
+            IpAddr::from_str("1.2.3.4").unwrap()
+        );
+        let req = prepare_request(API_GATEWAY_REST_GET_ROOT_ONEQUERY);
         assert_eq!(
             req.peer_addr().unwrap().ip(),
             IpAddr::from_str("1.2.3.4").unwrap()
         );
 
-        let reqjson: ApiGatewayV2 = serde_json::from_str(API_GATEWAY_V2_GET_REMOTE_IPV6).unwrap();
-        let req = actix_http::Request::try_from(reqjson).unwrap();
+        let req = prepare_request(API_GATEWAY_V2_GET_REMOTE_IPV6);
+        assert_eq!(
+            req.peer_addr().unwrap().ip(),
+            IpAddr::from_str("2404:6800:400a:80c::2004").unwrap()
+        );
+        let req = prepare_request(API_GATEWAY_REST_GET_REMOTE_IPV6);
         assert_eq!(
             req.peer_addr().unwrap().ip(),
             IpAddr::from_str("2404:6800:400a:80c::2004").unwrap()
@@ -337,22 +362,27 @@ mod tests {
     async fn test_form_post() {
         use actix_web::http::Method;
 
-        let reqjson: ApiGatewayV2 =
-            serde_json::from_str(API_GATEWAY_V2_POST_FORM_URLENCODED).unwrap();
-        let req = actix_http::Request::try_from(reqjson).unwrap();
+        let req = prepare_request(API_GATEWAY_V2_POST_FORM_URLENCODED);
+        assert_eq!(req.method(), Method::POST);
+        let req = prepare_request(API_GATEWAY_REST_POST_FORM_URLENCODED);
         assert_eq!(req.method(), Method::POST);
 
         // Base64 encoded
-        let reqjson: ApiGatewayV2 =
-            serde_json::from_str(API_GATEWAY_V2_POST_FORM_URLENCODED_B64).unwrap();
-        let req = actix_http::Request::try_from(reqjson).unwrap();
+        let req = prepare_request(API_GATEWAY_V2_POST_FORM_URLENCODED_B64);
+        assert_eq!(req.method(), Method::POST);
+        let req = prepare_request(API_GATEWAY_REST_POST_FORM_URLENCODED_B64);
         assert_eq!(req.method(), Method::POST);
     }
 
     #[test]
     fn test_parse_header() {
-        let reqjson: ApiGatewayV2 = serde_json::from_str(API_GATEWAY_V2_GET_ROOT_NOQUERY).unwrap();
-        let req = actix_http::Request::try_from(reqjson).unwrap();
+        let req = prepare_request(API_GATEWAY_V2_GET_ROOT_NOQUERY);
+        assert_eq!(req.head().headers.get("x-forwarded-port").unwrap(), &"443");
+        assert_eq!(
+            req.head().headers.get("x-forwarded-proto").unwrap(),
+            &"https"
+        );
+        let req = prepare_request(API_GATEWAY_REST_GET_ROOT_NOQUERY);
         assert_eq!(req.head().headers.get("x-forwarded-port").unwrap(), &"443");
         assert_eq!(
             req.head().headers.get("x-forwarded-proto").unwrap(),
@@ -362,16 +392,20 @@ mod tests {
 
     #[test]
     fn test_parse_cookies() {
-        let reqjson: ApiGatewayV2 = serde_json::from_str(API_GATEWAY_V2_GET_ROOT_NOQUERY).unwrap();
-        let req = actix_http::Request::try_from(reqjson).unwrap();
+        let req = prepare_request(API_GATEWAY_V2_GET_ROOT_NOQUERY);
         assert_eq!(req.head().headers.get("cookie"), None);
 
-        let reqjson: ApiGatewayV2 = serde_json::from_str(API_GATEWAY_V2_GET_ONE_COOKIE).unwrap();
-        let req = actix_http::Request::try_from(reqjson).unwrap();
+        let req = prepare_request(API_GATEWAY_V2_GET_ONE_COOKIE);
+        assert_eq!(req.head().headers.get("cookie").unwrap(), &"cookie1=value1");
+        let req = prepare_request(API_GATEWAY_REST_GET_ONE_COOKIE);
         assert_eq!(req.head().headers.get("cookie").unwrap(), &"cookie1=value1");
 
-        let reqjson: ApiGatewayV2 = serde_json::from_str(API_GATEWAY_V2_GET_TWO_COOKIES).unwrap();
-        let req = actix_http::Request::try_from(reqjson).unwrap();
+        let req = prepare_request(API_GATEWAY_V2_GET_TWO_COOKIES);
+        assert!(
+            req.head().headers.get("cookie").unwrap() == &"cookie2=value2; cookie1=value1"
+                || req.head().headers.get("cookie").unwrap() == &"cookie1=value1; cookie2=value2"
+        );
+        let req = prepare_request(API_GATEWAY_REST_GET_TWO_COOKIES);
         assert!(
             req.head().headers.get("cookie").unwrap() == &"cookie2=value2; cookie1=value1"
                 || req.head().headers.get("cookie").unwrap() == &"cookie1=value1; cookie2=value2"
