@@ -27,6 +27,9 @@ pub use rocket05::launch_rocket_on_lambda;
 #[cfg(feature = "hyper")]
 mod hyper014;
 #[cfg(feature = "hyper")]
+pub use hyper014::run_hyper_on_lambda;
+#[cfg(feature = "warp03")]
+#[deprecated(since = "0.1.8", note = "Use run_hyper_on_lambda()")]
 pub use hyper014::run_warp_on_lambda;
 #[cfg(feature = "warp03")]
 pub use warp;
